/*
 ftrig.h
  for fractional integer routines (alternative to floating-point)

  Created: 2008.07.15
  Revised: 2014.03.09

14.02.22-24 - POCC fixes
14.01.26 - added rfix16div
	fixed M68K ifix16div - now mimics fp overflow, ifixed not uint32.
13.11.02 - added ifract32mul
13.05.15 - added fractasin
13.05.10 - fixed POCC __inline _cntlz32
13.03.30
13.01.08 - ifix16div (handles overflow; no rounding)
11.06.24 - moved types to enviro64.h, added ufix16mul
*/

#ifndef __FTRIG_H
 #if defined(__POCC__)
  #pragma warn(push)
  #pragma warn(disable:2007 2096)
 #endif
/* returns IMIN(ifixed) on overflow */
 #define ifix16div(x,y) (ifixed)((ifixed)(x)*65536./(ifixed)(y))
/* fraction compensated rate division (unsigned, unrounded) */
 #define rfix16div(x,y,f) (ufixed)(((ufixed)(x)*65536.-(uint32)(f))/(ufixed)(y))
 #define ufract32div(x,y) (uint32)((uint32)(x)*4294967296./(uint32)(y))

 #ifdef env68k
  #pragma parameter __D0 ufract32mul(__D0, __D1)
  extern uint32 ufract32mul(uint32, uint32) = {
/*	MULU.L	D0,D0:D1 */	0x4C00, 0x1400
  };
  /* overflow will yield zero, as 0x100000000 would yield the same */
  #undef ufract32div
  #pragma parameter __D0 ufract32div(__D1,__D2)
  extern uint32 ufract32div(uint32, uint32) = {
/*	MOVEQ	#0,D0	*/	0x7000,
/*	DIVU.L	D2,D1:D0 */	0x4C42, 0x401
  };
  #pragma parameter __D0 ifract32mul(__D0, __D1)
  extern int32 ifract32mul(int32, int32) = {
/*	MULS.L	D0,D0:D1 */	0x4C00, 0x1C00
  };

  #pragma parameter __D0 ifract30mul(__D0, __D1)
  extern int32 ifract30mul(int32, int32) = {
/*	MULS.L	D0,D0:D1 */	0x4C00, 0x1C00,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180
  };
  #pragma parameter __D0 ufract30mul(__D0, __D1)
  extern uint32 ufract30mul(uint32, uint32) = {
/*	MULU.L	D0,D0:D1 */	0x4C00, 0x1400,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180
  };

  #pragma parameter __D1 ifract30div(__D0,__D2)
  extern int32 ifract30div(int32, int32) = {
/*	MOVEQ	#3,D1	*/	0x7203,
/*	AND.L	D0,D1	*/	0xC280,
/*	ROR.L	#2,D1	*/	0xE499,
/*	ASR.L	#2,D0	*/	0xE480,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	BPL.S	*+8	;@1 */	0x6A06,
/*	NEG.L	(A7)	*/	0x4497,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080,
/*@1	TDIVS.L	(A7),D2:D0 */	0x4C57, 0x802,
/*	DIVS.L	(A7)+,D2:D1 */	0x4C5F, 0x1C02
  };
  #pragma parameter __D1 ufract30div(__D0,__D2)
  extern uint32 ufract30div(uint32, uint32) = {
/*	MOVEQ	#3,D1	*/	0x7203,
/*	AND.L	D0,D1	*/	0xC280,
/*	ROR.L	#2,D1	*/	0xE499,
/*	LSR.L	#2,D0	*/	0xE488,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	TDIVU.L	D2,D2:D0 */	0x4C42, 2,
/*	DIVU.L	(A7)+,D2:D1 */	0x4C5F, 0x1402
  };

  #undef ifix16div
  #pragma parameter __D0 ifix16div(__D1,__D2)
  extern ifixed ifix16div(ifixed, ifixed) = {
/*	SWAP	D1	*/	0x4841,
/*	MOVE.L	D1,D0	*/	0x2001,
/*	EXT.L	D1	*/	0x48C1,
/*	CLR	D0	*/	0x4240,
/*	DIVS.L	D2,D1:D0 */	0x4C42, 0xC01,
/*	BVC.S	end	*/	0x6806,
/*	MOVE.L	#$80000000,D0*/	0x203C, 0x8000, 0
  };
  #undef rfix16div
  #pragma parameter __D0 rfix16div(__D0,__D2,__D1)
  extern uint32 rfix16div(uint32, uint32, uint32) = {
/*	MOVEA.L	D3,A0	*/	0x2043,
/*	SWAP	D0	*/	0x4840,
/*	MOVEQ	#0,D3	*/	0x7600,
/*	MOVE	D0,D3	*/	0x3600,
/*	CLR	D0	*/	0x4240,
/*	SUB.L	D1,D0	*/	0x9081,
/*	MOVEQ	#0,D1	*/	0x7200,
/*	SUBX.L	D1,D3	*/	0x9781,
/*	DIVU.L	D2,D3:D0 */	0x4C42, 0x403,
/*	BVC.S	@1	*/	0x6802,
/*	MOVEQ	#0,D0	*/	0x7000,
/*@1	MOVE.L	A0,D3	*/	0x2608
  };
  #pragma parameter __D0 ufix16mul(__D0, __D2)
  extern ufixed ufix16mul(ufixed, ufixed) = {
/*	MULU.L	D2,D2:D0 */	0x4C02, 0x402,
/*	ADD	D0,D0	*/	0xD040,
/*	MOVE	D2,D0	*/	0x3002,
/*	MOVEQ	#0,D2	*/	0x7400, 
/*	SWAP	D0	*/	0x4840,
/*	ADDX.L	D2,D0	*/	0xD182
  };
 #elif LONG_LONG
  #if __MRC__
   #define ufract32mul __mulhwu
  #elif defined(__POCC__) && envx86
   static __forceinline uint32 ufract32mul(uint32 a,
	uint32 b)
   {
	__asm
	{
		mov	eax,a
		mul	b
		mov	a,edx
	}
	return a;
   }
  #else
   #define ufract32mul(x,y) ((uint32)((uint64)(uint32)(x)*(uint32)(y)>>32))
  #endif

  #define ifract32mul(x,y) ((int32)((int64)(int32)(x)*(int32)(y)>>32))

  #define ifract30mul(x,y) ((int32)((int64)(int32)(x)*(int32)(y)>>30))
  #define ufract30mul(x,y) ((uint32)((uint64)(uint32)(x)*(uint32)(y)>>30))

  #define ifract30div(x,y) (int32)(((int64)(int32)(x)<<30)/(int32)(y))
  #define ufract30div(x,y) (uint32)(((uint64)(uint32)(x)<<30)/(uint32)(y))

  #define ufix16mul(x,y) (ufixed)((uint64)(ufixed)(x)*(ufixed)(y)+32768>>16)
 #else
  #define ufract32mul(x,y) (uint32)((uint32)(x)*(1/4294967296.)*(uint32)(y))

  #define ifract32mul(x,y) (int32)((int32)(x)*(1./4294967296.)*(int32)(y))

  #define ifract30mul(x,y) (int32)((int32)(x)*(1/1073741824.)*(int32)(y))
  #define ufract30mul(x,y) (uint32)((uint32)(x)*(1/1073741824.)*(uint32)(y))

  #define ifract30div(x,y) (int32)((int32)(x)*1073741824./(int32)(y))
  #define ufract30div(x,y) (uint32)((uint32)(x)*1073741824./(uint32)(y))

  #define ufix16mul(x,y) (ufixed)((ufixed)(x)/65536.*(ufixed)(y)+.5)
 #endif

/*
 vfix16mul/vfix16div
	For computing a fast, shifted multiply or divide, respectively.
	The result is clipped to the full ifixed range.
*/
ifixed vfix16mul(ifixed,
	ifixed);
ifixed vfix16div(ifixed,
	ifixed);

/*
 ihypot2/uhypot2/ihypot3/uhypot3
	These compute an RMS (root-mean-squareroot), to the formula,
		sqrt(a**2+b**2).
	They use fast integer computation, calling usqrt64.
	Common use: distance calculation, passing the coordinate differences.
  NOTE:
	Large enough parameters to uhypot3 can produce overflow.
*/
/* `sqrt(a**2+b**2)` or `sqrt((|a|/|b|)**2+1)*|b|` */
uint32 ihypot2(int32,
	int32);
uint32 uhypot2(uint32,
	uint32);
/* 2011.06.28 */
uint32 ihypot3(int32,
	int32,
	int32);
uint32 uhypot3(uint32,
	uint32,
	uint32);

/*
 usqrt/usqrt64
	These functions compute and integer square root, rounding.
	All modern implementations use a fast integer divide method.
*/
uint32 usqrt(uint32 n);
uint32 usqrt64(uint64 n);

/*
 rslopdiv
	Computes a fixed-point divide, rounding up for a div.-like ratio.
	Optionally, the initial value can be calculated for py if !=NULL.
  Result:
	The ratio, clipped to +/-0xFFFF.FFFF.
*/
int64 rslopdiv(int64 *px,
	int32 p0x,
	int32 p0y,
	int32 p1x,
	int32 p1y, /* must be >=p0y (sort the points) */
	uint16 ky); /* skip y (clipping) */
/*
 rincdiv
	Divides a range into the largest power of two for a uint64.
	It rounds up for a div.-like ratio.
	Optionally, the initial value can be calculated for py if !=NULL.
  Result:
	The incremental rate.  It may overflow in narrow y ranges.
*/
uint64 rincdiv(int64 *py,
	ifixed y0,
	ifixed y1, /* must be >y0 */
	uint16 ky); /* skip y (clipping) */

/*
 rarcmul
	Computes the formula w*sqrt(1-y*y), with y in +/-0x40000000 (+/-1.0).
	Values outside the valid range will result in clipped overflow.
	62-bit calculation ensures high accuracy.
  Result:
	The semicircle arc length, in [0,w/2].
*/
ufract rarcmul(ifract y,
	ufract w);

/*
 fastsin
	Computes a fast sine, given a binary angle.
	All values of [0,4.0) represent [0,2*pi).
  Accuracy note:
	The larger theta is in it's lower 30 bits, the less accurate the
       result (falling in between table entries with larger rise).
	All calculations are still accurate enough for 16 bit samples.
*/
ifract fastsin(ufract theta);
 #define fastcos(t) fastsin((t)+0x40000000u) /* phase shift */

/*
 ifractatan2/ufractatan2
	Computes an arctangent of (x/y), returning the binary angle.
	32-bit computation ensures fairly high accuracy.
*/
ufract ifractatan2(int32 x,
	int32 y);
ufract ufractatan2(uint32 x,
	uint32 y);
/*
 fractasin
	Computes a binary angle arcsine, using ifractatan2/usqrt64.
*/
/* if out of range (-1,1) returns min/max values */
ifract fractasin(ifract x);

/*
 cntlz32
	Computes a count of the leading zeroes in 32 bits, unsigned.
	This is a standard routine, utilizing native CPU instruction(s).
*/
int cntlz32(uint32 x);

/* built-in instructions */
 #ifdef env68k
  #pragma parameter __D0 _cntlz32(__D0)
  extern int _cntlz32(uint32) = {
/*	BFFFO	D0{0:0},D0 */	0xEDC0, 0
  };
  #define cntlz32 _cntlz32
 #elif defined(__MRC__)
  #define cntlz32(x) (int)__cntlzw((uint32)(x))
 #elif defined(__POCC__) && envx86
  static __forceinline int _cntlz32(uint32 v)
  {
	__asm
	{ /* default -1 for 32 result */
		or	eax,-1
		bsr	eax,v
		mov	v,31
		sub	v,eax
	}
	return v;
  }
  #define cntlz32(x) _cntlz32(x)
 #endif

 #if __POCC__
  #pragma warn(pop)
 #endif
 #define __FTRIG_H
#endif
