/*
 mblt.c

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

2014.02.24
  Took advantage of 32-bit div. result to maximize slopdiv ratio range.
(For horizontal lines, up half would not've shown up; now complete.)
2014.02.25
  Framed ovals!  Finally figured out the clipping mech.
rInitClip2 now a separate - but inlined - function.
2014.02.26
  Fixed framing outset, implemented round-rectangles (framed- broken).
2014.02.27
  Fixed framed round-rectangles, rewriting the initialization separately.
2014.03.05
  pixMap origin/bounds like QD, except x/y/w/h instead of 'rect16'.
  Fixed Asm slopdiv - the int64qset was treating a signed value unsigned.
  Got custom port clipping & origin working...
2014.03.06
  Separated the crux of line drawing into stdLine for an efficient rCLineList.
  rInitRoundRect: fixed incorrect use of clip-top when oy should be used.
2014.03.07
  With rgbToPenColor, colors are no longer expanded here (slightly faster).
  Translucency with alpha.
2014.03.08
  Universal support for all depths with fillx/invrx, using the same method
used in the M68K Asm version, but written from scratch in C.
Additional disadvantage of LSB: bswap required for string masks.
(The M68K version, however, used bitfield r/w - smaller.)
2014.03.09
  iTable 8bpp RGB translucency working, but slow.
- now sped up by building a direct table instead of passing the cSpec table;
at most 512 bytes, worth it.  Besides, od28-6 already builds one.
  Finally implemented region code... now, to get it to work...

2014.03.10
  Split from mgraf.c, into mblt and mgraphic.
  Then all the assembly into _mblt.
*/

#include "mgraf.h"

/* ---- code */

/* blt routines - see DEF_BLT_RTN in mgraf.h */

#if envx86
 extern DEF_BLT_RTN(blt_fill1);
 extern DEF_BLT_RTN(blt_fill2);
 extern DEF_BLT_RTN(blt_fill4);
 extern DEF_BLT_RTN(blt_fill8);
 extern DEF_BLT_RTN(blt_fill16);
 extern DEF_BLT_RTN(blt_fill32);
// extern DEF_BLT_RTN(blt_invr1);
// extern DEF_BLT_RTN(blt_invr2);
// extern DEF_BLT_RTN(blt_invr4);
#else
 #define fillx(h)	do{ uint32 *d, k, v; \
	o<<=h, d=(uint32*)a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int32)0x80000000>>(MIN(c,32)-1))>>o; \
	k=be32_native(k), v=f->vz, c+=o; \
	*d=(*d&~k)|(v&k); \
	if((c-=32)<=0) \
		return; \
	++d, o=c; \
	if(c>>=5) do *d++=v; while(--c); \
	if(o&=31) k=~(~0u>>o), k=be32_native(k), *d=(*d&~k)|(v&k); \
}while(0)

static DEF_BLT_RTN(blt_fill1)
	{fillx(0);}

static DEF_BLT_RTN(blt_fill2)
	{fillx(1);}

static DEF_BLT_RTN(blt_fill4)
	{fillx(2);}

/* -- only faster with small lines
static DEF_BLT_RTN(blt_fill4)
{
	int8		*d = a+(o>>1), v = f->vz;

	if (o&1)
		*d++ = (*d&0xF0)|(v&15), --c;
	if ((o=c)&2)
		*d++ = v;
	if (c>>=2)
		do
			*d++ = *d++ = v;
		while(--c);
	if (o&1)
		*d = (*d&15)|(v&0xF0);
}
*/

static DEF_BLT_RTN(blt_fill8)
{
	int8		*d = a+o, v = f->vz;

	if (c&1)
		*d++ = v;
	if (c>>=1)
		do
			*d++ = *d++ = v;
		while(--c);
}

static DEF_BLT_RTN(blt_fill16)
{
	int16		*d = (int16*)a+o, v = f->vz;

	if (c&1)
		*d++ = v;
	if (c>>=1)
		do
			*d++ = *d++ = v;
		while(--c);
}

static DEF_BLT_RTN(blt_fill32)
{
	int32		*d = (int32*)a+o, v = f->vz;

	if (c&1)
		*d++ = v;
	if (c>>=1)
		do
			*d++ = *d++ = v;
		while(--c);
}
#endif

/* inversion */

#define invrx(h)	do{ uint32 *d, k, v; \
	o<<=h, d=(uint32*)a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int32)0x80000000>>(MIN(c,32)-1))>>o; \
	k=be32_native(k), v=f->vz; \
	*d^=v&k, c+=o; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) do *d++^=v; while(--c); \
	if(o&=31) k=~(~0u>>o), *d^=be32_native(k)&v; \
}while(0)

static DEF_BLT_RTN(blt_invr1)
	{invrx(0);}

static DEF_BLT_RTN(blt_invr2)
	{invrx(1);}

static DEF_BLT_RTN(blt_invr4)
	{invrx(2);}

static DEF_BLT_RTN(blt_invr8)
{
	int16		v = f->vz;
	int8		*d = a+o;

	if ((int)d&1)
		*d++^= (int8)v, --c;
	o = c;
	if (c>>=1)
		do
			*(int16*)d^= v;
		while(d+=2,--c);
	if (o&=1)
		*d^= (int8)v;
}

static DEF_BLT_RTN(blt_invr16)
{
	int32		v = f->vz;
	int16		*d = (int16*)a+o;

	if ((int)d&1)
		*d++^= (int16)v, --c;
	o = c;
	if (c>>=1)
		do
			*(int32*)d^= v;
		while(d+=2,--c);
	if (o&=1)
		*d^= (int16)v;
}

static DEF_BLT_RTN(blt_invr32)
{
	int32		*d = (int32*)a+o, v = f->vz;

	if (c&1)
		*d++^= v;
	if (c>>=1)
		do
			*d^= v, d[1]^= v;
		while(d+=2,--c);
}

/* alpha fill */

static DEF_BLT_RTN(alphaFill8c) /* 8bpp color */
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->itb;
	uint8		*d = (uint8*)a+o;
	uint32		v = f->vz; /* direct */
	int		m = f->opa;

	do
	{
		uint32		a = dt[*d], r, g;

		r=a&f->opr,r=((((v&f->opr)-r)*m>>8)+r)&f->opr;
		g=a&f->opg,r+=((((v&f->opg)-g)*m>>8)+g)&f->opg;
		r+=a&=f->opb,*d=it[r+=(int32)((v&f->opb)-a)*m>>8];
	} while(++d,--c);
}

static DEF_BLT_RTN(alphaFill8g) /* 8bpp gray - *UNTESTED* */
{
	const uint8	*dt = f->dtb;
	const uint8	*it = f->itb;
	uint8		*d = (uint8*)a+o;
	uint32		v = f->vz; /* 16-bit */
	int		m = f->opa;

	do
	{
		uint32		a = dt[*d];

		*d = it[(a+((v-a)*m>>8))>>8];
	} while(++d,--c);
}

static DEF_BLT_RTN(alphaFill8f) /* 8bpp fixed (implied raw gray) */
{
	uint8		*d = (uint8*)a+o, v = f->vz, m = f->opa;

	do
	{
		uint8		a = *d;

		a&=f->opb, *d = a+((v-a)*m>>8);
	} while(++d,--c);
}

static DEF_BLT_RTN(alphaFill16)
{
	uint16		*d = (uint16*)a+o, v = f->vz;
	int		m = f->opa;

	do
	{
		uint16		a = *d, r, g;

		r=a&f->opr,r=((((v&f->opr)-r)*m>>8)+r)&f->opr;
		g=a&f->opg,r+=((((v&f->opg)-g)*m>>8)+g)&f->opg;
		r+=a&=f->opb,*d=r+=(int32)((v&f->opb)-a)*m>>8;
	} while(++d,--c);
}

static DEF_BLT_RTN(alphaFill32)
{
	uint32		*d = (uint32*)a+o, v = f->vz;
	int		m = f->opa;

	do
	{
		uint32		a = *d, r, g;
		/* highest must be pre-shifted due to 30bpp, etc. */
		r=a&f->opr,r=((((v&f->opr)>>8)-(r>>8))*m+r)&f->opr;
		g=a&f->opg,r+=((((v&f->opg)-g)*m>>8)+g)&f->opg;
		r+=a&=f->opb,*d=r+=(int32)((v&f->opb)-a)*m>>8;
	} while(++d,--c);
}

static bltRoutine	fillByDepth[] =
	{blt_fill1,blt_fill2,blt_fill4,blt_fill4,blt_fill8,blt_fill8,blt_fill8,blt_fill8};
static bltRoutine	invrByDepth[] =
	{blt_invr1,blt_invr2,blt_invr4,blt_invr4,blt_invr8,blt_invr8,blt_invr8,blt_invr8};

int rInitBlt(bltParam *f,
	grafPort *p, /* where the virtual origin/clip reside */
	int o,
	int c)
{
	pixMap		*m = p->map; /* where the actual origin/bounds reside */
	int		rb, g; /* red/blue component bit-size, green */
	/* clip */
	f->cl.top = MAX(p->origin.y,p->clip.top);
	f->cl.bottom = p->origin.y+m->h;
	if (f->cl.bottom>p->clip.bottom)
		f->cl.bottom = p->clip.bottom;
	if (c>f->cl.bottom)
		c = f->cl.bottom;
	if (c<=f->cl.top)
		return -1; /* clipped out-top */
	if (o>=f->cl.bottom)
		return -1; /* clipped out-bottom */
	/* set up the parameters */
	if ((f->p=lockPixels(m))==NULL)
		return -1; /* bad parameter */
	f->ox = p->origin.x+p->map->pOrigin.x;
	f->oy = p->origin.y+p->map->pOrigin.y;
	f->oz = m->pitch;
	/* caller should tell if object l/r-clips out of existence */
	f->cl.left = MAX(p->origin.x,p->clip.left);
	f->cl.right = p->origin.x+m->w;
	if (f->cl.right>p->clip.right)
		f->cl.right = p->clip.right;
	f->vz = p->pMode==pModeBic ? p->bkColor:p->fgColor;
	f->opa = p->alpha;
	f->opb = (1u<<(rb=m->minCmp))-1;
	if (m->pSize<9)
	{
		if (m->pTable)
		{
			if (!m->iTable && makeITables(m,m->bpp)<=0)
				return -2;
			f->itb = m->iTable;
			f->dtb = m->dTable;
			/* don't allow this bit to mess up masks */
			m->pType&=~pTypeLGreen;
		}
		//f->ih = 31-cntlz32(m->pSize);
		if (p->pMode==pModeXor)
		{
			f->rtn = invrByDepth[m->pSize-1];
			f->vz^= ((1u<<m->bpp)-1)*pSizeMasks[m->pSize-1];
		} else if (p->pMode==pModeBic || f->opa==0xFF || m->pSize<8)
			f->rtn = fillByDepth[m->pSize-1];
		else
		{
			if (m->pTable)
			{ /* separate: gray and color */
				f->vz&= ((1u<<m->bpp)-1);
				if (rb)
					f->rtn = alphaFill8c, f->vz = ((uint16*)f->dtb)[f->vz];
				else
					f->rtn = alphaFill8g, f->vz = ((uint8*)f->dtb)[f->vz];
			} else
				f->rtn = alphaFill8f, f->vz&= f->opb;
			goto opset;
		}
	} else if (m->pSize<17)
	{
		if (p->pMode==pModeXor)
			f->rtn = blt_invr16, f->vz^= ((1u<<m->bpp)-1)*0x10001;
		else if (p->pMode==pModeBic || f->opa==0xFF)
			f->rtn = blt_fill16;
		else
			{f->rtn = alphaFill16; goto opset;}
	} else if (p->pMode==pModeXor)
		f->rtn = blt_invr32, f->vz^= (2u<<(m->bpp-1))-1;
	else if (p->pMode==pModeBic || f->opa==0xFF)
		f->rtn = blt_fill32;
	else
	{
		f->rtn = alphaFill32;
	opset:	rb = m->minCmp, g = rb+(m->pType>>3&1); //pTypeLGreen
		f->opr = f->opb<<rb<<g;
		f->opg = ((1u<<g)-1)<<rb;
	}
	return rInitClip2(f,o,c,&f->a,&f->e);
}
