/*
 mgraphic.c

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

2014.03.10
  Split from mgraf.c, into mblt and mgraphic.
2014.03.11
  Got region code working, finally.  Rendering, too.
*/

#include <string.h>

#include "mgraf.h"
#include "ftrig.h"

/* ---- code */

static int rHalfOutRect(bltParam *f,
	rect2dq *b)
{
	ufixed		o;

	if (f->vx<0 || f->vy<0)
		f->vx =-1;
	else
	{
		if (!(f->vx|f->vy))
			return 0; /* hollow pen */
		/* adding any part of the fraction make ovals look bad */
		o = f->vx>>1&~0xFFFF;
		b->left = subu(b->left,o), b->right = addu(b->right,o);
		o = f->vy>>1&~0xFFFF;
		b->top = subu(b->top,o), b->bottom = addu(b->bottom,o);
	}
	if (b->bottom<=b->top)
		return 0;
	return 1;
}

/* renderers, starting with the trapezoid renderers
 rTrapezoid, rTriangle, stdLine:/rCLineList/rLineList
*/

#define initPenLine(f,p0,p1)	do{ \
	q2dsort2(p0,p1); /* sorting must come first, for p1y-p0y diff below */ \
	if ((f.vx|f.vy)<0) { /* non-pen */ \
		if (idim(p0.x,p1.x)>(ufixed)(p1.y-p0.y)) \
			f.vx=0, f.vy=0x10000; \
		else	f.vy=0, f.vx=0x10000; \
	} \
}while(0)

#define adjustHLine(f,o,c)	do{ \
	if (c>f.cl.right) c=f.cl.right; \
	if (o<f.cl.left) o=f.cl.left; \
	c-=o; \
	if ((o-=f.ox)<0) c+=o, o=0; \
}while(0)
/* two counts, adjusted with offset */
#define adjustHLine2(f,o,c,c2)	do{ \
	if (c>f.cl.right) c=f.cl.right; \
	if (c2>f.cl.right) c2=f.cl.right; \
	if (o<f.cl.left) o=f.cl.left; \
	c-=o,c2-=o; \
	if ((o-=f.ox)<0) c+=o,c2+=o,o=0; \
}while(0)

void rTrapezoid(trapezoid *f,
	pt2dq p0,
	pt2dq p1)
{
	int8		*a = f->prm.p, *e;
	int64		x0, x1, dx0, dx1;
	int32		c;
	int8		xr;
	/* clip bottom */
	if ((c=(p1.y>>16)-f->prm.cl.top)<0)
		c = 0;
	a+= f->prm.oz*(c+f->prm.cl.top-f->prm.oy);
	if (a<(e=f->prm.e))
		e = a;
	/* clip top */
	if ((a=f->prm.a)>=e)
		goto end;
	if ((c=f->prm.cl.top-(p0.y>>16))<0)
		c = 0;
	dx0 = rslopdiv(&x0,p0.x,p0.y,p1.x,p1.y,c);
	x1 = f->x1, dx1 = f->dx1;
	/* whether to reverse x0:x1 */
	xr = int64eq(x0,x1) ? int64ge(dx0,dx1):int64ge(x0,x1);
	if ((f->prm.vy=p0.y>>16)<f->prm.cl.top)
		f->prm.vy = f->prm.cl.top;
	do
	{ /* clip and blt line */
		int32		o;

		if (xr)
			o = int64hi(x1), c = int64hi(x0);
		else
			o = int64hi(x0), c = int64hi(x1);
		adjustHLine(f->prm,o,c);
		if (c>0)
			f->prm.rtn(o,c,a,&f->prm);
		(void)int64add(x0,dx0);
		(void)int64add(x1,dx1);
	} while(a+=f->prm.oz,a<e);
	f->x1 = x1;
end:	f->prm.a = e;
}

int rTriangle(grafPort *p,
	pt2dq p0,
	pt2dq p1,
	pt2dq p2)
{
	trapezoid	f; /* 2 trapezoids */
	int		c, bl, br;

	if (!p)
		return -1;
	/* sort */
	q2dsort2(p0,p1);
	q2dsort2(p0,p2);
	q2dsort2(p1,p2);
	if ((p0.x==p1.x && p0.y==p1.y) || (p1.x==p2.x && p1.y==p2.y)
			|| (c=rInitBlt(&f.prm,p,p0.y>>16,p2.y>>16))<0)
		return 0;
	/* clip horizontal bounds */
	if ((bl=br=p0.x)>p1.x)
		bl = p1.x;
	else
		br = p1.x;
	if (bl>p2.x)
		bl = p2.x;
	if ((bl>>=16)>f.prm.cl.left)
		f.prm.cl.left = bl;
	if (br<p2.x)
		br = p2.x;
	if ((br=br>>16)<f.prm.cl.right)
		f.prm.cl.right = br;
	if (f.prm.cl.right<=f.prm.cl.left)
		return 0;
	/* the only persistent slope */
	f.dx1 = rslopdiv(&f.x1,p0.x,p0.y,p2.x,p2.y,c);
	rTrapezoid(&f,p0,p1);
	rTrapezoid(&f,p1,p2);
	return 1;
}

static int stdLine(trapezoid *f,
	pt2dq p0,
	pt2dq p1,
	int32 ky)
{
	pt2dq		p2;
	ifixed		u;

	p2.y = addu(p1.y,f->prm.vy);
	p2.x = u = addu(p1.x,f->prm.vx);
	if (p0.x<p1.x)
	{ /* flat left: x1 = top of top-right to bottom-right */
		p2.x = p1.x, p1.x = u;
		f->dx1 = rslopdiv(&f->x1,addu(p0.x,f->prm.vx),p0.y,p1.x,p1.y,ky);
		/* clip for flat ends (error requires both l/r) */
		if ((u=p1.x>>16)<f->prm.cl.right)
			f->prm.cl.right = u;
		if ((u=p0.x>>16)>f->prm.cl.left)
			f->prm.cl.left = u;
	} else
	{ /* reverse: x1 = top of top-left to bottom-left */
		f->dx1 = rslopdiv(&f->x1,p0.x,p0.y,p1.x,p1.y,ky);
		p0.x = addu(p0.x,f->prm.vx);
		/* clip for flat ends (error requires both l/r) */
		if ((u=p0.x>>16)<f->prm.cl.right)
			f->prm.cl.right = u;
		if ((u=p1.x>>16)>f->prm.cl.left)
			f->prm.cl.left = u;
	}
	if (f->prm.cl.right<=f->prm.cl.left)
		return 0;
	p1.x = p0.x, p1.y = addu(p0.y,f->prm.vy);
	rTrapezoid(f,p0,p1);
	rTrapezoid(f,p1,p2);
	return 1;
}

static int rCLineList(trapezoid *f, /* must be initialized */
	const pt2dq *ep, /* must be valid */
	int ptc) /* must be >0 */
{
	rect32		cl = f->prm.cl;
	pt2dq		p0, p1;
	ifixed		pw = f->prm.vx, ph = f->prm.vy;
	ifixed		dx = pw>>1&~0xFFFF, dy = ph>>1&~0xFFFF;
	int32		o, c;
	int		r = 0;

	do
	{
		p0.x = subu(ep->x,dx), p0.y = subu(ep->y,dy);
		++ep;
		p1.x = subu(ep->x,dx), p1.y = subu(ep->y,dy);
		/* initialize the line's pen and clipping */
		f->prm.vx = pw, f->prm.vy = ph, f->prm.cl = cl;
		initPenLine(f->prm,p0,p1);
		if ((c=addu(p1.y,ph)>>16)>f->prm.cl.bottom)
			c = f->prm.cl.bottom;
		if (c<=f->prm.cl.top)
			continue; /* clipped out-top */
		if ((o=p0.y>>16)>=f->prm.cl.bottom)
			continue; /* clipped out-bottom */
		r|= stdLine(f,p0,p1,rInitClip2(&f->prm,o,c,&f->prm.a,&f->prm.e));
	} while(--ptc);
	/* QD does not force closure, and neither should this */
	return r;
}

int rLine(grafPort *p,
	pt2dq p0,
	pt2dq p1)
{ /* this takes advantage of l/r clipping - 2 trapezoids */
	trapezoid	f;
	int32		c;

	if (!p)
		return -1;
	f.prm.vx = p->penw, f.prm.vy = p->penh;
	if (!(f.prm.vx|f.prm.vy))
		return 0;
	initPenLine(f.prm,p0,p1);
	if ((c=rInitBlt(&f.prm,p,p0.y>>16,addu(p1.y,f.prm.vy)>>16))<0)
		return 0;
	return stdLine(&f,p0,p1,c);
}

/* regions */

/* For polygons, QD basically builds a sect-region.  This is optimal. */

SLRgn *newRgn(void)
{
	SLRgn		*rgn;

	if ((rgn=malloc(OFFSET(SLRgn,x)))==NULL)
		return NULL;
	rgn->bBox.top = rgn->bBox.left = IMAX(rgn->bBox.left);
	rgn->bBox.bottom = rgn->bBox.right = IMIN(rgn->bBox.right);
	rgn->size = OFFSET(SLRgn,x);
	return rgn;
}

int rgnSectLine(SLRgn **rrgn,
	int16 ctop,
	int16 cbot,
	pt2dq p0,
	pt2dq p1)
{
	SLRgn		*rgn;
	const int16	*f, *s, *e;
	int16		*a;
	int64		x, dx;
	int32		o, c, z, newa, newb;
	int32		y0i, y1i, x0i, x1i;

	if (!rrgn)
		return -1;
	if ((rgn=*rrgn)==NULL)
		if ((rgn=newRgn())==NULL)
			return -2;
		else
			*rrgn = rgn;
	if ((z=rgn->size)<OFFSET(SLRgn,x)
			|| rgn->bBox.bottom<=rgn->bBox.top)
		rgn->size = z = OFFSET(SLRgn,x);
	q2dsort2(p0,p1);
	if ((y0i=p0.y>>16)<ctop)
		y0i = ctop;
	if ((y1i=p1.y>>16)>cbot)
		y1i = cbot;
	if ((c=y1i-y0i)<=0)
		return 0; /* no addition */
	dx = rslopdiv(&x,p0.x,p0.y,p1.x,p1.y,y1i-(p0.y>>16)); /* from end to work backward */
	o = z;
	/* if empty, match here to reduce code */
	if (rgn->bBox.bottom<=rgn->bBox.top)
		rgn->bBox.top = rgn->bBox.bottom = y0i;
	if ((newa=rgn->bBox.top-y0i)>0)
		z+= newa*(2*sizeof(int16)); /* creation of new lines above */
	if ((newb=y1i-rgn->bBox.bottom)>0)
		z+= newb*(2*sizeof(int16)); /* creation of new lines below */
	z+= c*sizeof(int16); /* inserted positions */
	if ((rgn=realloc(rgn,z))==NULL)
		return -2;
	*rrgn = rgn;
	/* working from end renders ability to perform insertion on one buffer */
	f = rgn->x; /* cannot run before this (would be erroneous) */
	e = a = (int16*)(z+(int8*)rgn); /* new write */
	s = (int16*)(o+(int8*)rgn); /* old read */
	if (p0.x>p1.x)
		o = p1.x, c = p0.x;
	else
		o = p0.x, c = p1.x;
	if ((x0i=o>>16)<rgn->bBox.left)
		rgn->bBox.left = x0i;
	if ((x1i=c>>16)>rgn->bBox.right)
		rgn->bBox.right = x1i;
	if (newb<0) /* copy only */
		do
		{
			*--a = *--s; /* end taken unconditionally */
			if (a<=f || s<f)
				return -1;
			while(s>f && s[-1]<=*s)
				*--a = *--s;
		} while(++newb);
	else if (newb)
		do
		{
			if (a<f)
				return -1;
			int64sub(x,dx);
			o = int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			*--a = IMAX(*a), *--a = o, *--a = IMIN(*a);
		} while(--newb);
	if ((c=MIN(rgn->bBox.bottom,y1i)-MAX(rgn->bBox.top,y0i))>0)
		do
		{ /* tested: this appears to work with all inserted x values */
			if (a<f || s<=f)
				return -1;
			int64sub(x,dx);
			o = int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			/* endings require neg.-space, and xs can be IMIN/IMAX */
			/* be sure not to remove the start/end when compacting */
			*--a = *--s; /* end taken unconditionally */
			while(s[-1]>o)
				*--a = *--s;
			/* coming before less or same */
			if (a<=f || s<f)
				return -1;
			if (*--s!=o || s<=f || s[-1]>o) /* compact */
				*--a = o, *--a = *s;
			while(s>f && s[-1]<=*s)
				*--a = *--s;
		} while(--c);
	if (newa<0) /* copy only */
		do
		{
			*--a = *--s; /* end taken unconditionally */
			if (a<=f || s<f)
				return -1;
			while(s>f && s[-1]<=*s)
				*--a = *--s;
		} while(++newa);
	else if (newa)
		do
		{
			if (a<=f)
				return -1;
			int64sub(x,dx);
			o = int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			*--a = IMAX(*a), *--a = o, *--a = IMIN(*a);
		} while(--newa);
	if (a<f)
		return -1;
	rgn->size = (z=(int8*)e-(int8*)a)+OFFSET(SLRgn,x);
	if (a!=f)
		memcpy(rgn->x,a,z); /* compacting made the region smaller */
	/* success!  can now finish the bounds. */
	if (y0i<rgn->bBox.top)
		rgn->bBox.top = y0i;
	if (y1i>rgn->bBox.bottom)
		rgn->bBox.bottom = y1i;
	return 1;
}

static void outsetRect(rect16 *b,
	int x,
	int y)
{
	int32		v;

	if ((v=b->top-x)<IMIN(b->top))
		v = IMIN(b->top);
	b->top = v;
	if ((v=b->left-x)<IMIN(b->left))
		v = IMIN(b->left);
	b->left = v;
	if ((v=b->bottom+y)>IMAX(b->bottom))
		v = IMAX(b->bottom);
	b->bottom = v;
	if ((v=b->right+y)>IMAX(b->left))
		v = IMAX(b->left);
	b->right = v;
}

int rRegion(grafPort *p,
	const SLRgn *rgn)
{
	bltParam	f;
	rect16		b;
	const int16	*s, *e;
	int32		o, c;

	if (!p || !rgn || (c=rgn->size)<OFFSET(SLRgn,x) || c&1)
		return -1;
	s = rgn->x;
	e = (int16*)(rgn->size+(int8*)rgn);
	if (e<=s || (o=p->penw|p->penh)==0)
		return 0;
	b = rgn->bBox;
	if (o>=0)
		outsetRect(&b,p->penw>>17,p->penh>>17);
	if ((c=rInitBlt(&f,p,b.top,b.bottom))<0
			|| b.left>=f.cl.right
			|| b.right<=f.cl.left)
		return 0;
	--e;
	if (c) /* skip clipped-out lines */
		do
		{
			if (s>=e || s[1]<*s)
				return -1;
			for(++s; *s>=s[-1]; ++s)
				;
		} while(--c);
	do
	{
		if (s>=e || s[1]<*s)
			return -1;
		do
		{
			++s;
			if ((c=s[1])<(o=*s))
				break;
			++s;
			adjustHLine(f,o,c);
			if (c>0)
				f.rtn(o,c,f.a,&f);
		} while(s[1]>=*s);
	} while(++s,f.a+=f.oz,f.a<f.e);
	return 0;
}

/* polygons */

int rPolygon(grafPort *p,
	const pt2dq *ep,
	int count)
{
	trapezoid	f;
	SLRgn		*rgn;
	const pt2dq	*pt;
	int32		c, bl, br;

	if (!p || (pt=ep)==NULL || (c=count)<2)
		return -1;
	f.prm.vx = p->penw, f.prm.vy = p->penh;
	if (!(f.prm.vx|f.prm.vy))
		return 0; /* hollow pen */
	/* cl temporarily used to build the polygon's fixed-point bounds */
	bl = f.prm.cl.top = IMAX(ifixed);
	br = f.prm.cl.bottom = IMIN(ifixed);
	do
	{
		if (bl>pt->x)
			bl = pt->x;
		if (br<pt->x)
			br = pt->x;
		if (f.prm.cl.top>pt->y)
			f.prm.cl.top = pt->y;
		if (f.prm.cl.bottom<pt->y)
			f.prm.cl.bottom = pt->y;
	} while(++pt,--c);
	if (f.prm.vx>=0)
	{ /* frame expansion */
		br+= f.prm.vx, f.prm.cl.bottom+= f.prm.vy;
		c = f.prm.vx>>1&~0xFFFF;
		bl-= c, br+= c;
		c = f.prm.vy>>1&~0xFFFF;
		f.prm.cl.top = subu(f.prm.cl.top,c);
		f.prm.cl.bottom = addu(f.prm.cl.bottom,c);
	}
	if ((c=rInitBlt(&f.prm,p,f.prm.cl.top>>16,f.prm.cl.bottom>>16))<0)
		return 0;
	if ((bl>>=16)>f.prm.cl.left)
		f.prm.cl.left = bl;
	if ((br>>=16)<f.prm.cl.right)
		f.prm.cl.right = br;
	if (f.prm.cl.right<=f.prm.cl.left)
		return 0;
	--count;
	if (f.prm.vx>=0) /* frame */
		return rCLineList(&f,ep,count);
	if (count<2)
		return 0; /* no area */
	rgn = NULL, pt = ep;
	do
		if (rgnSectLine(&rgn,f.prm.cl.top,f.prm.cl.bottom,*pt,pt[1])<0)
			goto bad;
	while(++pt,--count);
	/* area painting requires automatic close */
	if (rgnSectLine(&rgn,f.prm.cl.top,f.prm.cl.bottom,*pt,*ep)<0)
	{
	bad:	if (rgn)
			free(rgn);
		return -1;
	}
	rRegion(p,rgn);
	if (rgn)
		free(rgn);
	return 1;
}

/* rectangle-parameter functions
  rOval, rRect, rRoundRect
*/

int rOval(grafPort *p,
	const rect2dq *ov)
{
	bltParam	f;
	rect2dq	b; /* outer-bounds */
	int8		*ia, *ie; /* interior markers (outline case) */
	uint64		dy0, dy1;
	int64		y0, y1;
	ufixed		w;
	ifixed		cx;
	int32		c, o;

	if (!ov || !p)
		return -1;
	b = *ov, f.vx = p->penw, f.vy = p->penh;
	/* outset must be calculated before clipping */
	if (!rHalfOutRect(&f,&b) || (c=rInitBlt(&f,p,b.top>>16,b.bottom>>16))<0)
		return 0;
	/* rarely, some narrow fractions/large ratios manage to jutt a lot */
	if ((o=b.left>>16)>f.cl.left)
		f.cl.left = o;
	if ((o=(b.right-1)>>16)<f.cl.right)
		f.cl.right = o; /* to not -1 this will result in full-width edge pixels */
	if (f.cl.right<=f.cl.left)
		return 0;
	/* cx +/- w~/(1-(y/h)**2) */
	w = b.right-b.left;
	cx = b.left+(w>>1)+(w&1); /* any round not toward hollow left */
	/* set up y increment & position */
	dy0 = rincdiv(&y0,b.top,b.bottom,c);
	int64subhl(y0,0x40000000,0);
	int64xset(dy1,0);
	if (w<=f.vx || (f.vx=w-f.vx*2)+0u>w) /* for non-pen case, >-1u is impossible */
		ia = ie = (int8*)-1;
	else
	{
		ifixed		it = b.top+f.vy, ib = b.bottom-f.vy;

		c = rInitClip2(&f,it>>16,(ib-1)>>16,&ia,&ie);
		if (ia>=ie || ia>=f.e)
			ia = (int8*)-1; /* clipped out or hollow */
		else
		{
			dy1 = rincdiv(&y1,it,ib,c);
			int64subhl(y1,0x40000000,0);
		}
	}
	if ((f.vy=b.top>>16)<f.cl.top)
		f.vy = f.cl.top;
	do
	{
		o = rarcmul(int64hi(y0),w);
		c = addu(cx,o)>>16, o = subu(cx,o)>>16;
		if (f.a>=ia)
			if (f.a>=ie)
				ia = (int8*)-1;
			else
			{ /* hollow interior */
				ufixed	i = rarcmul(int64hi(y1),f.vx);
				int32	ic;

				ic = subu(cx,i)>>16;
				adjustHLine(f,o,ic);
				if (ic>0)
					f.rtn(o,ic,f.a,&f);
				o = (addu(cx,i)-1)>>16;
				(void)int64add(y1,dy1);
			}
		adjustHLine(f,o,c);
		if (c>0)
			f.rtn(o,c,f.a,&f);
		(void)int64add(y0,dy0);
	} while(f.a+=f.oz,f.a<f.e);
	return 1;
}

int rRect(grafPort *p,
	const rect2dq *rc)
{
	bltParam	f;
	rect2dq	b; /* outer-bounds */
	int8		*ia, *ie; /* interior markers (outline case) */
	ufixed		w;
	int32		o, c, ic1, io, ic2;

	if (!rc || !p)
		return -1;
	b = *rc, f.vx = p->penw, f.vy = p->penh;
	/* outset must be calculated before clipping */
	if (!rHalfOutRect(&f,&b) || rInitBlt(&f,p,b.top>>16,b.bottom>>16)<0)
		return 0;
	o = b.left>>16, ic2 = c = b.right>>16;
	w = b.right-b.left, ic1 = io = 0;
	if (w<=f.vx || w-f.vx*2u>w)
		ia = ie = (int8*)-1;
	else
	{
		ifixed		it = b.top+f.vy, ib = b.bottom-f.vy;
		/* `(ib-1)>>16` produces thicker line in zero-frac case */
		rInitClip2(&f,it>>16,ib>>16,&ia,&ie);
		if (ia>=ie || ia>=f.e)
			ia = (int8*)-1; /* clipped out or hollow */
		else
		{
			ic1 = (b.left+f.vx)>>16;
			io = (b.right-f.vx)>>16;
			adjustHLine(f,io,ic2);
		}
	}
	/* rRect the fastest, with all the clipping done before the loop */
	adjustHLine2(f,o,c,ic1);
	if (c<=0)
		return 0;
	if ((f.vy=b.top>>16)<f.cl.top)
		f.vy = f.cl.top;
	do
	{
		if (f.a>=ia)
			if (f.a>=ie)
				ia = (int8*)-1;
			else
			{ /* hollow interior */
				if (ic1>0)
					f.rtn(o,ic1,f.a,&f);
				if (ic2>0)
					f.rtn(io,ic2,f.a,&f);
				continue;
			}
		f.rtn(o,c,f.a,&f);
	} while(f.a+=f.oz,f.a<f.e);
	return 1;
}

/* Fast clipping is the hard part with fixed-point round rectangles;
 must get the full-width vertical range in integral coordinates.
   Also, no matter how you slice it (no pun intended), the lower part
 fraction & skip must be calculated separately.
   All of this is complex enough to separate it all into its own function.
*/
typedef struct RRect
{
	const int8	*ua, *ue; /* full-width markers */
	int64		dy, y0, y1; /* y-inc, top of rect-rounding positions */
	ufixed		rw, mw; /* round & mid-width; rarcmul(y,rw)+mw */
} RRect;

static void rInitRoundRect(RRect *r,
	const bltParam *f,
	int32 ky,
	ufixed w,
	ifixed bt,
	ifixed bb,
	ufixed rx,
	ufixed ry,
	ufixed oy)
{
	int32		uy0, uy1;
	uint16		uy1f;

	if ((rx+=rx)>w)
		rx = w; /* x-collapsed */
	r->rw = rx, r->mw = (w-rx)>>1;
	uy0 = addu(bt,ry)>>16, uy1 = subu(bb,ry);
	uy1f = (uint16)uy1, uy1>>= 16;
	ry+= ry;
	if (uy1<=uy0)
		r->ua = r->ue = (int8*)-1, ry = bb-bt; /* y-collapsed */
	else
	{
		r->ua = f->p+f->oz*(uy0-f->oy);
		r->ue = f->p+f->oz*(uy1-f->oy);
	}
	/* set up y increment & position */
	if (!ry)
	{
		int64xset(r->dy,0);
		r->y0 = r->y1 = r->dy;
	} else
	{
		r->dy = rincdiv(&r->y0,bt,bt+ry,ky);
		int64subhl(r->y0,0x40000000,0);
		r->y1 = r->dy; int64shr(r->y1,16);
		if ((ky=f->cl.top-uy1)<0)
			ky = 0;
		lineFrac(ky,ky,uy1f);
		int64mulu32(r->y1,ky);
	}
}

int rRoundRect(grafPort *p,
	const rect2dq *rc,
	ifixed rx,
	ifixed ry)
{
	bltParam	f;
	rect2dq	b; /* outer-bounds */
	RRect		rr, rri; /* exterior & interior (outline case) */
	int8		*ia, *ie; /* interior markers */
	ufixed		w;
	ifixed		cx;
	int32		c, o;

	if (rx<0)
		rx = 0;
	if (ry<0)
		ry = 0;
	if (!(rx|ry))
		return rRect(p,rc);
	if (!rc || !p)
		return -1;
	b = *rc, f.vx = p->penw, f.vy = p->penh;
	/* outset must be calculated before clipping */
	if (!rHalfOutRect(&f,&b) || (c=rInitBlt(&f,p,b.top>>16,b.bottom>>16))<0)
		return 0;
	/* rarely, some narrow fractions/large ratios manage to jutt a lot */
	if ((o=b.left>>16)>f.cl.left)
		f.cl.left = o;
	if ((o=(b.right-1)>>16)<f.cl.right)
		f.cl.right = o; /* to not -1 this will result in full-width edge pixels */
	if (f.cl.right<=f.cl.left)
		return 0;
	w = b.right-b.left;
	cx = b.left+(w>>1)+(w&1); /* any frac round not toward hollow left */

	rInitRoundRect(&rr,&f,c,w,b.top,b.bottom,rx,ry,p->origin.y);
	if (w<=f.vx || (f.vx=w-f.vx*2)+0u>w)
		ia = ie = (int8*)-1;
	else
	{
		ifixed		it = b.top+f.vy, ib = b.bottom-f.vy;

		c = rInitClip2(&f,it>>16,ib>>16,&ia,&ie);
		if (ia>=ie || ia>=f.e)
			ia = (int8*)-1; /* clipped out or hollow */
		else
			rInitRoundRect(&rri,&f,c,f.vx,it,ib,rx,ry,p->origin.y);
	}
	if ((f.vy=b.top>>16)<f.cl.top)
		f.vy = f.cl.top;
	do
	{
		if (f.a>=rr.ua)
			if (f.a>=rr.ue)
				rr.ua = (int8*)-1, rr.y0 = rr.y1;
			else
			{
				o = b.left>>16, c = b.right>>16;
				goto uo;
			}
		o = rarcmul(int64hi(rr.y0),rr.rw)+rr.mw;
		(void)int64add(rr.y0,rr.dy);
		c = addu(cx,o)>>16, o = subu(cx,o)>>16;
	uo:	if (f.a>=ia)
			if (f.a>=ie)
				ia = (int8*)-1;
			else
			{ /* hollow interior */
				ufixed	i;
				int32	ic;

				if (f.a>=rri.ua)
					if (f.a>=rri.ue)
						rri.ua = (int8*)-1, rri.y0 = rri.y1;
					else
					{
						i = rri.rw>>1;
						goto ui;
					}
				i = rarcmul(int64hi(rri.y0),rri.rw);
				(void)int64add(rri.y0,rri.dy);
			ui:	i+= rri.mw, ic = subu(cx,i)>>16;
				adjustHLine(f,o,ic);
				if (ic>0)
					f.rtn(o,ic,f.a,&f);
				o = (addu(cx,i)-1)>>16;
			}
		adjustHLine(f,o,c);
		if (c>0)
			f.rtn(o,c,f.a,&f);
	} while(f.a+=f.oz,f.a<f.e);
	return 1;
}
