/*
 mkvtab.c
  Virtual Key Map, with initial basis on Darwin/Frameworks-Events.h

  Avoid mapping keys that sound a little like the proper key.
 We don't want more official/modern keys going to the wrong function.
 e.g, it would be bad, just any easy key causing a POWER/RESET hit.
*/

#include <enviro64.h>
#include "mevent.h"

#define BL KBL		/* blank */
#define RD 66		/* redundant - must be properly updated */

/* we don't */
const uint8	mkvtab_win[227] = 
{
	BL, /* 0 - reserved/null */
	BL, /* VK_LBUTTON */
	BL, /* VK_RBUTTON */
	BL, /* VK_CANCEL */
	BL, /* 4 - VK_MBUTTON */
	BL, /* VK_XBUTTON1 */
	BL, /* VK_XBUTTON2 */
	BL, /* undef */
	51, /* 8 - VK_BACK (backspace) */
	48,
	BL, /* undef */
	BL,
	143, /* 12 - VK_CLEAR */
	36, /* VK_RETURN */
	BL,
	BL,
	RD, /* 16 - VK_SHIFT */
	RD, /* VK_CONTROL */
	RD, /* VK_MENU */
	113, /* VK_PAUSE - F15, pulse-only on pre-Windows 7 */
	57, /* 20 - VK_CAPITAL */
	104, /* VK_KANA/HANGEUL/HANGUL */
	BL, /* undef */
	136, /* VK_JUNJA */
	137, /* 24 - VK_FINAL */
	138, /* VK_HANJA/KANJI */
	BL, /* undef */
	53, /* VK_ESCAPE */
	BL, /* 28 - VK_CONVERT */
	BL, /* VK_NONCONVERT */
	137, /* VK_ACCEPT */
	BL, /* VK_MODECHANGE */
	49, /* 32 - VK_SPACE */
	116, /* VK_PRIOR */
	121, /* VK_NEXT */
	119, /* VK_END */
	115, /* 36 - VK_HOME */
	123, /* VK_LEFT */
	126, /* VK_UP */
	124, /* VK_RIGHT */
	125, /* 40 - VK_DOWN */
	128, /* VK_SELECT */
	BL, /* VK_PRINT */
	141, /* VK_EXECUTE */
	105, /* 44 - VK_SNAPSHOT - F13, pulse-only */
	114, /* VK_INSERT */
	117, /* VK_DELETE */
	112, /* VK_HELP */
	29, /* 48 - VK_0 */
	18, /* VK_1 */
	19, /* VK_2 */
	20, /* VK_3 */
	21, /* 52 - VK_4 */
	23, /* VK_5 */
	22, /* VK_6 */
	26, /* VK_7 */
	28, /* 56 - VK_8 */
	25, /* VK_9 */
	BL,
	BL,
	BL, /* 60 - undef */
	BL,
	BL,
	BL,
	BL, /* 64 - undef */
	0, /* VK_A */
	11, /* VK_B */
	8, /* VK_C */
	2, /* 68 - VK_D */
	14, /* VK_E */
	3, /* VK_F */
	5, /* VK_G */
	4, /* 72 - VK_H */
	34, /* VK_I */
	38, /* VK_J */
	40, /* VK_K */
	37, /* 76 - VK_L */
	46, /* VK_M */
	45, /* VK_N */
	31, /* VK_O */
	35, /* 80 - VK_P */
	12, /* VK_Q */
	15, /* VK_R */
	1, /* VK_S */
	17, /* 84 - VK_T */
	32, /* VK_U */
	9, /* VK_V */
	13, /* VK_W */
	7, /* 88 - VK_X */
	16, /* VK_Y */
	6, /* VK_Z */
	63, /* VK_LWIN */
	54, /* 92 - VK_RWIN */
	58, /* VK_APPS - to option */
	BL, /* undef */
	BL, /* VK_SLEEP - not wise to map this, I take it */
	82, /* 96 - VK_NUMPAD0 */
	83, /* VK_NUMPAD1 */
	84, /* VK_NUMPAD2 */
	85, /* VK_NUMPAD3 */
	86, /* 100 - VK_NUMPAD4 */
	87, /* VK_NUMPAD5 */
	88, /* VK_NUMPAD6 */
	89, /* VK_NUMPAD7 */
	91, /* 104 - VK_NUMPAD8 */
	92, /* VK_NUMPAD9 */
	67, /* VK_MULTIPLY/kp* */
	69, /* VK_ADD/kp+ */
	81, /* 108 - VK_SEPARATOR --> KP= */
	78, /* VK_SUBTRACT/kp- */
	65, /* VK_DECIMAL/kp. */
	75, /* VK_DIVIDE/kpslash */
	122, /* 112 - VK_F1 */
	120, /* VK_F2 */
	99, /* VK_F3 */
	118, /* VK_F4 */
	96, /* 116 - VK_F5 */
	97, /* VK_F6 */
	98, /* VK_F7 */
	100, /* VK_F8 */
	101, /* 120 - VK_F9 */
	109, /* VK_F10 */
	103, /* VK_F11 */
	111, /* VK_F12 */
	105, /* 124 - VK_F13 - these keys may not work the same */
	107, /* VK_F14 */
	113, /* VK_F15 - must be done manually */
	106, /* VK_F16 */
	64, /* 128 - VK_F17 */
	79, /* VK_F18 */
	80, /* VK_F19 */
	90, /* VK_F20 */
	132, /* 132 - VK_F21 */
	133, /* VK_F22 */
	134, /* VK_F23 */
	135, /* VK_F24 */
	BL, /* 136 - undef */
	BL,
	BL,
	BL,
	BL, /* 140 */
	BL,
	BL,
	BL,
	71, /* 144 - async:VK_NUMLOCK */
	107, /* async:VK_SCROLL */
	BL, /* VK_OEM_NEC_EQUAL/FJ_JISHO */
	BL, /* VK_OEM_FJ_MASSHOU */
	BL, /* 148 - VK_OEM_FJ_TOUROKU */
	BL, /* VK_OEM_FJ_LOYA */
	BL, /* VK_OEM_FJ_ROYA */
	BL,
	BL, /* 152 - undef */
	BL,
	BL,
	BL,
	BL, /* 156 - undef */
	BL,
	BL,
	BL,
	56, /* 160 - VK_LSHIFT */
	60, /* VK_RSHIFT */
	59, /* VK_LCONTROL */
	62, /* VK_RCONTROL */
	55, /* 164 - VK_LMENU */
	61, /* VK_RMENU */
	153, /* VK_BROWSER_BACK */
	154, /* VK_BROWSER_FORWARD */
	155, /* 168 - VK_BROWSER_REFRESH */
	156, /* VK_BROWSER_STOP */
	157, /* VK_BROWSER_SEARCH */
	158, /* VK_BROWSER_FAVORITES */
	159, /* 172 - VK_BROWSER_HOME */
	74, /* VK_VOLUME_MUTE */
	73, /* VK_VOLUME_DOWN */
	72, /* VK_VOLUME_UP */
	152, /* 176 - VK_MEDIA_NEXT_TRACK */
	151, /* VK_MEDIA_PREV_TRACK */
	150, /* VK_MEDIA_STOP */
	149, /* VK_MEDIA_PLAY_PAUSE */
	BL, /* 180 - VK_LAUNCH_MAIL */
	BL, /* VK_LAUNCH_MEDIA_SELECT */
	BL, /* VK_LAUNCH_APP1 */
	BL, /* VK_LAUNCH_APP2 */
	BL, /* 184 - undef */
	BL,
	41, /* VK_OEM_1 */
	24, /* VK_OEM_PLUS */
	43, /* 188 - VK_OEM_COMMA */
	27, /* VK_OEM_MINUS */
	47, /* VK_OEM_PERIOD */
	44, /* VK_OEM_2/slash */
	50, /* 192 - VK_OEM_3/grave */
	BL, /* undef */
	BL,
	BL,
	BL, /* 196 - undef */
	BL,
	BL,
	BL,
	BL, /* 200 - undef */
	BL,
	BL,
	BL,
	BL, /* 204 */
	BL,
	BL,
	BL,
	BL, /* 208 */
	BL,
	BL,
	BL,
	BL, /* 212 */
	BL,
	BL,
	BL,
	BL, /* 216 */
	BL,
	BL,
	33, /* VK_OEM_4/lbracket */
	42, /* 220 - VK_OEM_5/bkslash */
	30, /* VK_OEM_6/rbracket */
	39, /* VK_OEM_7/squote */
	BL, /* VK_OEM_8 */
	BL, /* 224 - undef */
	BL, /* VK_OEM_AX */
	129, /* VK_OEM_102/less */
};	/* VK_ICO_HELP */
	/* 228 - VK_ICO_00 */
	/* VK_PROCESSKEY */
	/* VK_ICO_CLEAR */
	/* VK_PACKET */
	/* 232 - undef */
	/* VK_OEM_RESET */
	/* VK_OEM_JUMP */
	/* VK_OEM_PA1 */
	/* 236 - VK_OEM_PA2 */
	/* VK_OEM_PA3 */
	/* VK_OEM_WSCTRL */
	/* VK_OEM_CUSEL */
	/* 240 - VK_OEM_ATTN */
	/* VK_OEM_FINISH */
	/* VK_OEM_COPY */
	/* VK_OEM_AUTO */
	/* 244 - VK_OEM_ENLW */
	/* VK_OEM_BACKTAB */
	/* VK_ATTN */
	/* VK_CRSEL */
	/* 248 - VK_EXSEL */
	/* VK_EREOF */
	/* VK_PLAY */
	/* VK_ZOOM */
	/* 252 - VK_NONAME */
	/* VK_PA1 */
	/* VK_OEM_CLEAR */
